from typing import BinaryIO, List, Union

from aidkit_client._endpoints.constants import Constants
from aidkit_client._endpoints.models import (
    FrameworkType,
    ImageModelContextSpecs,
    ListMLModelResponse,
    MLModelCreationRequestForSchema,
    MLModelResponse,
    MLModelVersionResponse,
    TextModelContextSpecs,
)
from aidkit_client.aidkit_api import HTTPService
from aidkit_client.exceptions import ResourceWithIdNotFoundError


class MLModelsAPI:
    api: HTTPService

    def __init__(self, api: HTTPService):
        self.api = api

    async def get_all(self) -> List[MLModelResponse]:
        result = await self.api.get(path=Constants.ML_MODELS_PATH, parameters=None)
        return ListMLModelResponse(
            **result.body_dict_or_error("Failed to retrieve MLModels.")
        ).items

    async def get(self, name: str) -> MLModelResponse:
        result = await self.api.get(path=f"{Constants.ML_MODELS_PATH}/{name}", parameters=None)
        if result.is_not_found:
            raise ResourceWithIdNotFoundError(f"MLModel with name {name} not found")
        return MLModelResponse(
            **result.body_dict_or_error(f"Error fetching MLModel with name {name}. ")
        )

    async def update(self, name: str, new_name: str) -> MLModelResponse:
        result = await self.api.patch(
            path=f"{Constants.ML_MODELS_PATH}/{name}",
            parameters=None,
            body={"name": new_name},
        )
        if result.is_bad:
            raise ResourceWithIdNotFoundError(f"Could not update MLModel name with {name}")
        return MLModelResponse(
            **result.body_dict_or_error(f"Error patching MLModel with name {name}.")
        )

    async def upload_model_version(
        self, model_name: str, model_version_name: str, model_file_content: BinaryIO, file_name: str
    ) -> MLModelVersionResponse:

        result = await self.api.post_multipart_data(
            path=f"{Constants.ML_MODELS_PATH}/{model_name}/versions",
            data={"name": model_version_name},
            files={"model": (f"{file_name}", model_file_content)},
        )
        if result.is_bad:
            raise ResourceWithIdNotFoundError(
                f"Could not upload version '{model_version_name}' of MLModel '{model_name}'."
            )
        return MLModelVersionResponse(
            **result.body_dict_or_error(
                f"Error uploading version '{model_version_name}' of MLModel '{model_name}'."
            )
        )

    async def create_model(
        self,
        model_name: str,
        model_version_name: str,
        framework: FrameworkType,
        context: Union[ImageModelContextSpecs, TextModelContextSpecs],
        model_file_content: BinaryIO,
    ) -> MLModelResponse:

        # Schema cannot be changed because it is derived from the request in
        # the MLModelPresenter, so that we pass empty bytes to the non-used
        # model argument.
        model_creation_request = MLModelCreationRequestForSchema(
            model=bytes(),
            name=model_name,
            version_name=model_version_name,
            context=context,
            framework=framework,
        )
        # model_creation_request is used for encoding the context as json but cannot be passed
        # on directly to the server because the model file binary cannot be encoded in json
        result = await self.api.post_multipart_data(
            path=f"{Constants.ML_MODELS_PATH}",
            data={
                "name": model_creation_request.name,
                "version_name": model_creation_request.version_name,
                "context": model_creation_request.context.json(),
                "framework": model_creation_request.framework.value,
            },
            files={"model": model_file_content},
        )

        return MLModelResponse(
            **result.body_dict_or_error(
                f"Error creating MLModel '{model_name}'"
                f" or uploading model version '{model_version_name}'."
            )
        )

    async def delete(self, name: str) -> bool:
        response = await self.api.delete(path=f"{Constants.ML_MODELS_PATH}/{name}")
        return response.is_success
